//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "TMainForm.h"
#include "TDebuggingForm.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TMainForm *MainForm;
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsPropertyName( TStringList* property_name_strings, AnsiString property_name )
{
for (int i=0;i<property_name_strings->Count;i++) if (property_name_strings->Strings[i]==property_name) return true;
return false;
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsInterfaceName( TStringList* Interface_name_strings, AnsiString Interface_name )
{
for (int i=0;i<Interface_name_strings->Count;i++) if (Interface_name_strings->Strings[i]==Interface_name) return true;
return false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ImportInterfaces( AnsiString pFileName )
{
DebuggingForm->Caption = "Importing DirectX Interfaces.... debugging output";

ImportRow = 0;

// convert it as required
fInterfaceNames = new TStringList();
fInterfaceDescriptions = new TStringList();
fPropertyNames = new TStringList();
fPropertyDescriptions = new TStringList();
fPropertyReturnTypes = new TStringList();
fPropertyReturnValues = new TStringList();
fParameterTypes = new TStringList();
fParameterNames = new TStringList();
fParameterModifiers = new TStringList();   //can be "*" or "FAR *"
fParameterDescriptions = new TStringList();
fFlagNames = new TStringList();
fFlagDescriptions = new TStringList();

// Immediately following the 'Interfaces' keyword is a list of the different
// Interfaces defined inside the file.
// By scanning rows for Interface names we build a list of the various
// Interface that we are going to parse.
// When a Interface name is repeated it signifies that the first Interface
// definition is following
ImportRow++;

Dbg( ">>> determining the various Interfaces included in this file <<<");
bool found_first_repeated_Interface = false;
do
    {
    // check the current row to see if it contains a unique structure name
    Dbg( "checking row for a unique Interface name" );
    if (IsInterfaceName(fInterfaceNames, ImportString))
        {
        Dbg( "found a repetition of a Interface name: '" + ImportString + "'... assuming the relevant structure list is finished." );
        found_first_repeated_Interface = true;
        break;  // continue at end of this for loop
        }
    // if we didn't find a unique Interface name
    if (!found_first_repeated_Interface)
        {
        // check that the row conforms to the structure naming rules
        if (IsValidInterfaceName(ImportString))
            {
            Dbg( "found a valid Interface name: '" + ImportString + "'... adding to internal structures list" );
            fInterfaceNames->Add( ImportString );
            ImportRow++;
            if (ImportRow == ImportStrings->Count) ERROR( "scanned all rows but did not find a Interface name repeated after the 'Interface included in this file' block...");
            }
        else ERROR( "scanned all rows but did not find a Interface name repeated after the 'Structures included in this file' block...");
        }
    }

while (!found_first_repeated_Interface);

// since we now know how many interfaces we need to parse, attempt to parse that many interfaces
for (int interface_index=0;interface_index<fInterfaceNames->Count;interface_index++)
    {
    Dbg( ">>> parsing Interface definition #" + IntToStr(interface_index) + " <<<" );

    // clear internal storage
    //TStringList* TempFunctionNames = new TStringList();
	//TStringList* TempFunctionDescriptions = new TStringList();
	//TStringList* TempFunctionReturnValueTypes = new TStringList();
	TStringList* TempParameterTypes = new TStringList();
	TStringList* TempParameterModifiers = new TStringList();
	TStringList* TempParameterNames = new TStringList();
	TStringList* TempParameterDescriptions = new TStringList();
    TStringList* TempFlagNames = new TStringList();
	TStringList* TempFlagDescriptions = new TStringList();
	// since we already know the names, check that the definition is appropriate but otherwise
    // ignore this row.
    Dbg( "Checking Interface name to make sure that a definition is following");
    if (ImportString!=fInterfaceNames->Strings[interface_index]) ERROR( "Did not find the expected Interface name: '" + fInterfaceNames->Strings[interface_index] + "'" );
    Dbg( "name is valid for Interface: '" + fInterfaceNames->Strings[interface_index] + "'..." );
    Dbg( "... Assuming that a Interface definition for '" + fInterfaceNames->Strings[interface_index] + "' is following" );
    Msg( "Parsing Interface: '" + fInterfaceNames->Strings[interface_index] + "'" );
    ImportRow++;

    // simply add the Interface description
    Dbg( "Assuming row #" + IntToStr( ImportRow ) + " is a description for Interface: '" + fInterfaceNames->Strings[interface_index] + "'" );
    fInterfaceDescriptions->Add( ImportString );
    ImportRow++;



    // read in the properties
    Dbg( "checking row for a property token..." );
    if (!IsSingleToken(ImportString)) ERROR( "property list not found for '" + fInterfaceNames->Strings[interface_index] + "'" );
    Dbg( "... property list found ..." );
    fPropertyNames->Add( ImportString );
    ImportRow++;

    // the following rows may occur anywhere.... hence we loop till we find the end of the 'struct' definition
    bool end_of_properties = false;
    Dbg( "parsing properties..." );
    do
        {
        // skip out when we find the end-token  IDirect3DRMArray::GetSize
        if ((ImportString.Pos("::") > 0) && IsSingleToken(ImportString))
            {
            Dbg( "Found the end of the properties list." );
            end_of_properties = true;
            if (!IsSingleToken(ImportString)) ImportRow++ ;
            continue;
            }
   		fPropertyNames->Add( ImportString );
        ImportRow++;
        }
    while (!end_of_properties);


    // we now process the property descriptions and their parameters.
    for (int property_index=0;property_index<fPropertyNames->Count;property_index++)
    	{
        // "IDirect3DRMArray::GetSize"
        if (fPropertyNames->Strings[property_index] == "AddRef") continue;
        if (fPropertyNames->Strings[property_index] == "QueryInterface") continue;
        if (fPropertyNames->Strings[property_index] == "Release") continue;
 	   	Dbg( "evaluating for property " + fPropertyNames->Strings[property_index] ); //TStringList
 	   	Dbg( "expecting a '::' keyword... comparing..." );
	    if (ImportString.Pos("::")==0) ERROR( "Did not find the '::' keyword as expected." );
 	   	Dbg( "row contained a valid '::' keyword" );
 	   	ImportRow++;
		// now we have a description for the Property
		if (IsSingleToken(ImportString))
        	{
        	//error, not a description
        	ERROR( "Did not find Property Description as expected." );
		   	//ImportRow++;
        	}
    	else
    		{
       	 	fPropertyDescriptions->Add( ImportString );
 	   		ImportRow++;
       		}
        // next ... is DWORD GetSize( );
        // we get the "DWORD" as PropertyReturnType
        if (ImportString.Pos("(") == 0) ERROR ("Did not find PropertyReturnType line as expected");
        fPropertyReturnTypes->Add(ImportString.SubString(1,ImportString.Pos(" ")));
  		Dbg( "Adding PropertyReturnTypes" + ImportString.SubString(1,ImportString.Pos(" ")));
        bool parameter_list_found = false;
        do
        	{
        	if (ImportString.Pos(");") > 0) continue;  // and expect return value
     	   	ImportRow++;
            parameter_list_found = true;
         	if (ImportString.Pos(");") > 0) continue;  // and expect return value
			//LPDIRECT3DRMDEVICE lplpD3DRMDevice,
			//LPDIRECT3DRMDEVICE *lplpD3DRMDevice,
			//LPDIRECT3DRMDEVICE * lplpD3DRMDevice,
			//LPDIRECT3DRMDEVICE FAR *lplpD3DRMDevice,
			//LPDIRECT3DRMDEVICE FAR * lplpD3DRMDevice
           //ERROR ("Deliberate Stop - we have a list ") ;

	        if (ImportString.Pos("*") != 0)
            	// found "*", ParameterName is ltrim of the rest of the substring
                // this takes the FAR into account when we create the parameter modifyer
                // which can be "*" or "FAR *"

            	{
                int x = ImportString.Pos("*");
                int y = ImportString.Pos(",");
                int z = ImportString.Pos(" ");
                int w = ImportString.Length();
                if (y>0)
                	y = w - x - 1 ;//chop off the ","
                else
                    y = w - x;
                TempParameterNames->Add(ImportString.SubString(x+1,y).TrimLeft()); //ltrim this
                TempParameterModifiers->Add(ImportString.SubString(z+1,x-z));
                TempParameterTypes->Add(ImportString.SubString(1,z-1));
    			Dbg( "adding name'" +ImportString.SubString(x+1,y-1));
    			Dbg( "adding typemodifiername'" +ImportString.SubString(x+1,y)+"'"+ ImportString.SubString(z+1,x-z)+"'"+ImportString.SubString(1,z-1)+"'");
    			//Dbg( "adding name'" +ImportString.SubString(x,y)+"'"+ ImportString.SubString(z,x-z)+"'"+ImportString.SubString(1,z)+"'");
                }
            else
            	{
                int y = ImportString.Pos(",");
                int z = ImportString.Pos(" ");
                int w = ImportString.Length();
                if (y>0)
                	y = w - z - 1 ;//chop off the ","
                else
                    y = w - z;
                TempParameterNames->Add(ImportString.SubString(z+1,y).TrimLeft()); //some trimming required
                TempParameterTypes->Add(ImportString.SubString(1,z-1));
    			Dbg( "adding type, and name"+ImportString );
    			Dbg( "adding type and name'"+ImportString.SubString(z+1,y)+"'"+ ImportString.SubString(1,z-1)+"'");
                }
            }


        while (ImportString.Pos(");") == 0);
        ImportRow++;
        //if (ImportString=="Parameters")
        //	{
        //    ImportRow++; //read the parameters line
        //    Dbg("Skipping Parameters Line");
        //    }
        //ERROR ("Deliberate Stop - no list or list complete ") ;
        //we now expect return values for the property or descriptions of the parameters

      	// we can now parse the descriptions of the various parameters of the interface.
     	// note that some parameters also contain descriptions of various flags.
      	// note also that other parameters contain "something and somethingelse" which
     	// we need to distinguish between.
		if (parameter_list_found)
        	{
       		if (ImportString=="Parameters")
        		{
            	ImportRow++; //read the parameters line
            	Dbg("Skipping Parameters Line");
            	}
   			Dbg( "attempting to parse the descriptions for the various interface parameters" );
   			bool checking_last_parameter_flags = false;
   			for (int parameter_index=0;parameter_index<TempParameterNames->Count;parameter_index++)
   		 	   {
   		   		Dbg( "parsing Property parameter  #" + IntToStr(parameter_index) );
   		    	// check for the premature end of the parameter descriptions
   		    	if (IsKeyword(ImportString,"QuickInfo"))
   	       		 	{
   		    	    // special case: if we're checking for flags for the last member, and the check failed
   	        		// do not check for "this and that" rules.
   	        		if (!checking_last_parameter_flags) ERROR( "found a 'QuickInfo' section before all parameter descriptions were defined." );
   	        		Dbg( "found a 'QuickInfo' section after checking all parameter descriptions. Assuming parameter descriptions are completed" );
   	        		break;
   	        		}
    	    	if (IsKeyword(ImportString,"Remarks"))
    	    	    {
    	    	    // special case: if we're checking for flags for the last member, and the check failed
    	    	    // do not check for "this and that" rules.
    	    	    if (!checking_last_parameter_flags) ERROR( "found a 'Remarks' section before all parameter descriptions were defined." );
    	    	    Dbg( "found a 'Remarks' section after checking all parameter descriptions. Assuming parameter descriptions are completed" );
    	    	    break;
    	    	    }
                // if the current line contains the name of the appropriate member, ignore the member name &
    	    	// store the description that follows immediately after this row.
    	    	Dbg( "expecting parameter name: '" + TempParameterNames->Strings[parameter_index] + "'... checking to make sure that row is a valid parameter name" );
    	    	if (ImportString==TempParameterNames->Strings[parameter_index])
    	    	    {
    	    	    Dbg( "found a valid parameter name: '" + TempParameterNames->Strings[parameter_index] + "'" );
    	    	    Msg( "Parsing Interface Parameter: '" + fInterfaceNames->Strings[interface_index] + "::" + TempParameterNames->Strings[parameter_index] + "'" );
    	        	ImportRow++;

       	     		Dbg( "assuming row #" + IntToStr(ImportRow) + " as the description for parameter: '" + TempParameterNames->Strings[parameter_index] + "'" );
    	        	TempParameterDescriptions->Add( EncodeValue("InterfaceIndex",interface_index) + ImportString );
    	        	ImportRow++;

    	        	// check if there are any flags following the member description
   	         		Dbg( "checking to see if flags are declared following the parameter description" );
                	if (IsFlag(ImportString))
             		   	{
             		   	Dbg( "found a flag rule immdediately following the parameter description." );
             		   	do
             		  		{
                	   		Dbg( "adding a parameter flag: '" + ImportString + "'" );
             		       	Msg( "Found Interface Parameter Flag: '" + fParameterNames->Strings[parameter_index] + "::" + TempParameterNames->Strings[parameter_index] + "'" + "::" + ImportString );
             		       	TempFlagNames->Add( ImportString);
             		       	// check for flag descriptions declared on the same line as the flag name
                	       	int temp_pos = ImportString.Pos(" ");
              		 	   	if (temp_pos!=0)
                	        	{
                	        	Dbg( "the parameter flag row also contains the relevant flag description" );
                	  	 		AnsiString flag_description = ImportString.SubString(temp_pos,ImportString.Length()-temp_pos).TrimLeft();
                	        	Dbg( "adding the parameter flag description: '" + flag_description + "'" );
                	        	TempFlagDescriptions->Add(EncodeValue("FlagInterfaceIndex",interface_index) + EncodeValue("FlagParameterIndex",parameter_index) + flag_description);
                	        	}
                	       	else
                	        	// check just to make sure that the flags do not have descriptions with them...
                	        	if (IsFlag(ImportString))
                	            	{
                	            	Dbg( "the parameter flag description is assumed to be on the next row..." );
                	            	ImportRow++;
                	            	Dbg("adding a parameter flag description: '" + ImportString + "'" );
                	            	TempFlagDescriptions->Add(EncodeValue("FlagInterfaceIndex",interface_index) + EncodeValue("FlagParameterIndex",parameter_index) + ImportString);
                	            	ImportRow++;
                	            	}
                	        	else
                	            	ERROR( "a flag was encountered followed immediately by another flag with no descriptions !" );
                	        }
                	    while (IsFlag(ImportString));
                		}
            		else
                		Dbg( "there are no flags following the parameter description" );
            		}
        		else
                	// it's not a simple parameter, where one was expected.... maybe it's a case of "this and that".... or "this, that, and something else"
            		{
            		Dbg( "did not find a parameter name where expected... checking 'this and that' rules..." );
            		// if there is another parameter name following this one....
            		if (parameter_index<TempParameterNames->Count-1) // room for this and that
                		{
                		Dbg( "...checking 'this and that' rules");
                		AnsiString first_parameter_name;
                		AnsiString second_parameter_name;
                		if (IsThisAndThat(TempParameterNames,ImportString,&first_parameter_name,&second_parameter_name))
                		    {
                		    Dbg( "the row fits the rules for a 'this and that' type description" );
                		    ImportRow++;    // skip past the "this and that" row
                		    Dbg( "assuming row #" + IntToStr(ImportRow) + " is the description for parameters: '" + first_parameter_name + "' and '" + second_parameter_name + "'" );
                		    TempParameterDescriptions->Add( EncodeValue("InterfaceIndex",interface_index) + ImportString );
                		    TempParameterDescriptions->Add( EncodeValue("InterfaceIndex",interface_index) + ImportString );
                		    ImportRow++;
                		    parameter_index++;
                		    }
                		else
                		    {
                		    Dbg( "the row did not fit the 'this and that' rule. Checking 'this, that, and something else' rule..." );
                		    TStringList* temp_parameter_names = new TStringList();
                		    if (IsThisAndThatAndSomethingElse(TempParameterNames,ImportString, temp_parameter_names ))
                		        {
               		 	        Dbg( "the row fits the rules for a 'this, and that and something else' type description" );
                		        ImportRow++;    // skip past the "this, that and something else" row
                		        for (int i=0;i<temp_parameter_names->Count;i++)
                		            {
                		            Dbg( "adding parameter description for parameter: '" + temp_parameter_names->Strings[i] + "'" );
                		            TempParameterDescriptions->Add( EncodeValue("InterfaceIndex",interface_index) + ImportString );
                		            }
                		        if (temp_parameter_names->Count>0) parameter_index += temp_parameter_names->Count-1;
                		        ImportRow++;
                		        }
               			    else
                		        {
                	            Dbg( "Found the following parameter names when checking 'this, that, and something else' rule" );
                	        	for (int i=0;i<temp_parameter_names->Count;i++) Dbg( "parameter #" + IntToStr(i) + "=='" + temp_parameter_names->Strings[i] + "'" );
             		 			temp_parameter_names->Clear();
             					delete temp_parameter_names;
             		 			ERROR( "Expected but did not find a particular parameter name & description: '" + fInterfaceNames->Strings[interface_index] + "::" + TempParameterNames->Strings[parameter_index] + "'" );
								}
 							temp_parameter_names->Clear();
           	    	     	delete temp_parameter_names;
                	    	}
                		}
            		else
            	    	ERROR( "Expected but did not find a particular parameter name & description: '" + fInterfaceNames->Strings[interface_index] + "::" + TempParameterNames->Strings[parameter_index] + "'" );
            		}
        		}
               	if (ImportString=="Return Values")
               		{
               		Dbg("found 'Return Values'");
               		ImportRow ++ ;
               		// read a string into database
               		fPropertyReturnValues->Add(ImportString);
               		Dbg("'Property Return Values' are" + ImportString);
               		ImportRow ++ ;
               		}

            }

    	// we now expect either a "Remarks", "Return values", new Interface or a "QuickInfo" section
    	// check just to make sure
    	// if it is a 'Remarks' section, simply skip it till we find a "QuickInfo" section.

        if (ImportString=="Return Values")
               {
               Dbg("found 'Return Values'");
               ImportRow ++ ;
               // read a string into database
               fPropertyReturnValues->Add(ImportString);
               Dbg("'Property Return Values' are" + ImportString);
               ImportRow ++ ;
               }
     	if (ImportString=="Remarks")
    	    while ((ImportString!="QuickInfo") && !IsInterfaceName(fInterfaceNames,ImportString) )
    	        {
    	        ImportRow++;
    	        if (ImportRow>=ImportStrings->Count)
    	            {
    	            Dbg( "Could not parse the import file '" + pFileName + "'.\n" +
                               "Found a 'Remarks' section but did not find the 'QuickInfo' section following" +
                                ImportString );
    	            Dbg( "ERROR LINE: " + ImportString );
    	            DebuggingForm->Show();
    	            DebugLabel->Caption = "";
    	            return;
        	        }
        	    }

		//    Dbg( "Checking if the keyword 'QuickInfo' exists.");
		//    Dbg( "string being checked == '" + ImportString + "'" );
    if (ImportString=="QuickInfo")
        {
     	ImportRow += 5;
        Dbg( " a 'QuickInfo' section" );
        }
        }
    // yey. the structure definition has been successfully parsed.
    Dbg( ">>>>>> Definition of structure '" + fInterfaceNames->Strings[interface_index] + "' was completed successfully! <<<<<<" );

    //    Dbg( "Clearing structure internal representations" );
    int i;
    for (i=0;i<TempFlagDescriptions->Count;i++) fFlagDescriptions->Add( TempFlagDescriptions->Strings[i] );
    for (i=0;i<TempFlagNames->Count;i++) fFlagNames->Add( TempFlagNames->Strings[i] );
    for (i=0;i<TempParameterDescriptions->Count;i++) fParameterDescriptions->Add( TempParameterDescriptions->Strings[i] );
    for (i=0;i<TempParameterTypes->Count;i++) fParameterTypes->Add( TempParameterTypes->Strings[i] );
    for (i=0;i<TempParameterModifiers->Count;i++) fParameterModifiers->Add( TempParameterModifiers->Strings[i] );
    for (i=0;i<TempParameterNames->Count;i++) fParameterNames->Add( TempParameterNames->Strings[i] );

    TempFlagDescriptions->Clear();
    TempFlagNames->Clear();
    TempParameterDescriptions->Clear();
    TempParameterTypes->Clear();
    TempParameterNames->Clear();
    TempParameterModifiers->Clear();

    delete TempFlagDescriptions;
    delete TempFlagNames;
    delete TempParameterDescriptions;
    delete TempParameterTypes;
    delete TempParameterNames;
    delete TempParameterModifiers;
    }

Dbg( IntToStr(fInterfaceNames->Count) + " Interfaces were imported SUCCESSFULLY!" );
Msg( IntToStr(fInterfaceNames->Count) + " Interfaces were imported SUCCESSFULLY!" );

}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//brians code ends here Hi Hi Hi Hi Hi Hi Hi Hi Hi Hi Hey Hey Hey STOP NOW !!
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
    : TForm(Owner)
{
fImportStrings = new TStringList();
fImportRow = 0;
fLinked = true;
}
//---------------------------------------------------------------------------
__fastcall TMainForm::~TMainForm()
{
delete fImportStrings;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//brians code back up further Hi Hi Hi Hi Hi Hi Hi Hi Hi Hi Hey Hey Hey STOP NOW !!
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void __fastcall TMainForm::ImportFile( AnsiString pFileName )
{
DebuggingForm->Caption = "Importing DirectX data?Files.... debugging output";
DebuggingForm->ListBox->Items->Clear();

// load the file into a temporary TStringList
ImportStrings->Clear();

// then copy it to another TStringList so that there are no blank rows
TStringList* temp_ImportStrings = new TStringList();
temp_ImportStrings->LoadFromFile( pFileName );
for (int i=0;i<temp_ImportStrings->Count;i++) if (!temp_ImportStrings->Strings[i].TrimLeft().TrimRight().IsEmpty()) ImportStrings->Add(temp_ImportStrings->Strings[i].TrimLeft().TrimRight());
delete temp_ImportStrings;

Dbg( "Importing file: '" + pFileName + "'.... " );
Msg( "Importing file: '" + pFileName + "'.... " );
ImportRow = 0;

// scan through each of the 'strings' in the input file & parse as we go.

// check the file to make sure it contains structures
Dbg( ">>> checking for 'Import File Type' definition (ie. Structures, Interfaces, etc. <<<");
if (IsKeyword(ImportString,"Structures"))
    {
    Dbg( "Found the 'Import File Type' keyword...");
    Dbg( "... '" + pFileName + "' may be a 'Structures' definition file.... attempting to parse...");
    ImportRow++;
    ImportStructures( pFileName );
    }
else if (IsKeyword(ImportString,"Interfaces"))
    {
    Dbg( "Found the 'Import File Type' keyword...");
    Dbg( "... '" + pFileName + "' may be a 'Interfaces' definition file.... attempting to parse...");
    ImportRow++;
    ImportInterfaces( pFileName );
    }
else
    ERROR( "There was no file type definition at the top of the file..." );
}
//---------------------------------------------------------------------------
void        __fastcall  TMainForm::FSetImportRow( int row )
{
if (row<ImportStrings->Count) Dbg( "Row #" + IntToStr(row) + ": '" + ImportStrings->Strings[row] + "'" );
fImportRow = row;
}
//---------------------------------------------------------------------------
AnsiString  __fastcall  TMainForm::FGetImportString()
{
if (ImportRow<ImportStrings->Count) return ImportStrings->Strings[ImportRow];
return "";
}
//---------------------------------------------------------------------------
AnsiString  __fastcall  TMainForm::RemoveHungarianNotation( AnsiString hungarian_string )
{
char* p = hungarian_string.c_str();
int i=0;
for (i=0;i<hungarian_string.Length();i++)
    {
    if ((*p <= 'A') || (*p >= 'Z')) break;
    p++;
    }
return hungarian_string.SubString( i, hungarian_string.Length()-i );
}
//---------------------------------------------------------------------------
AnsiString  __fastcall  TMainForm::EncodeValue( AnsiString code, int value )
{
// build an encoded string
return "::[" + code + "," + IntToStr(value) + "]::";
}
//---------------------------------------------------------------------------
int         __fastcall  TMainForm::DecodeValue( AnsiString code, AnsiString encoded_string, AnsiString* decoded_string )
{
//Dbg( ">>> Inside DecodeValue().. Code is: '" + code + "'" );
//Dbg( ">>> Inside DecodeValue().. Encoded string is: '" + encoded_string + "'" );
while (true)
    {
    // determine if this supplied string has been encoded
    int start_pos = encoded_string.Pos("::[");
    int end_pos = encoded_string.Pos("]::");
    if ((start_pos==0) || (end_pos==0))
        {
//        if (start_pos==0) Dbg( "could not find a start token" );
//        if (end_pos==0) Dbg( "could not find an end token" );
        *decoded_string = encoded_string;
        return -1;    // exit if not encoded
        }

    // if (a) the current encoded_section contains the valid code
    // (b) a comma is inside the encoded string
//    Dbg("encoded section has been found. extracting");
    AnsiString encoded_section = encoded_string.SubString( start_pos+3, end_pos-start_pos-3 );
//    Dbg("encoded_section == '" + encoded_section + "'" );
//    Dbg("searching for a comma");
    int comma_pos = encoded_section.Pos(",");
    if ((encoded_section.Pos(code)==1) && (comma_pos != 0))
        // return the encoded value
        {
//        Dbg("Found relevant code name + comma in encoded section");
        AnsiString value_string = encoded_section.SubString( comma_pos+1, encoded_section.Length() - comma_pos );
//        Dbg("value_string == '" + value_string + "'" );
        encoded_string.Delete( start_pos, end_pos-start_pos+3 );
        *decoded_string = encoded_string;
        return StrToInt( value_string );
        }

    // delete the encoded section and continue to look for more encoded sections
//    Dbg("did not find appropriate encoded section... temporarily deleting subsection of string");
    encoded_string.Delete( start_pos, end_pos-start_pos+3 );
//    Dbg("after deleting subsection of string, the encoded_string == '" + encoded_string + "'" );
    }
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsFlag( AnsiString string )
{
// if there is a space, maybe its a flag with a description on the same line
int temp_pos = string.Pos(" ");
if (temp_pos!=0) string = string.SubString(0,temp_pos-1);
return (string.AnsiCompare( UpperCase(string)) == 0);
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsThisAndThat( TStringList* known_items, AnsiString string, AnsiString* this_string, AnsiString* that_string )
{
int and_pos = string.Pos(" and ");
if (and_pos==0) return false;
*this_string = string.SubString(1,and_pos-1);
Dbg("inside IsThisAndThat():: this_string == '" + *this_string + "'" );
*that_string = string.SubString( and_pos+5, ImportString.Length() - and_pos+1 );
Dbg("inside IsThisAndThat():: that_string == '" + *that_string + "'" );
return (IsMemberName(known_items,*this_string) && IsMemberName(known_items,*that_string));
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsThisAndThatAndSomethingElse( TStringList* known_items, AnsiString string, TStringList* result_strings )
{
result_strings->Clear();
//Dbg("inside IsThisAndThatAndSomethingElse():: before checking for comma, string == '" + string + "'" );
int comma_pos = string.Pos(",");
while (comma_pos!=0)
    {
    result_strings->Add( string.SubString(1,comma_pos-1) );
//Dbg("inside IsThisAndThatAndSomethingElse():: found a comma... added substring == '" + string.SubString(1,comma_pos-1) + "'" );
    string.Delete(1,comma_pos+1);
//    Dbg("inside IsThisAndThatAndSomethingElse():: after deletion, string == '" + string + "'" );
//Dbg("inside IsThisAndThatAndSomethingElse():: before checking for comma, string == '" + string + "'" );
    comma_pos = string.Pos(",");
    }
//Dbg("inside IsThisAndThatAndSomethingElse():: string == '" + string + "'" );
//Dbg("inside IsThisAndThatAndSomethingElse():: checking for 'and'" );
int and_pos = string.Pos("and ");
if (and_pos!=0)
    {
//    Dbg("inside IsThisAndThatAndSomethingElse():: found an 'and'... adding substring '" + string.SubString(and_pos+4,string.Length() - (and_pos+3)) + "'" );
    result_strings->Add( string.SubString(and_pos+4,string.Length() - (and_pos+3)) );
    }
bool result = false;
for (int i=0;i<result_strings->Count;i++) result |= IsMemberName(known_items,result_strings->Strings[i]);
return result;
}
//---------------------------------------------------------------------------
void    __fastcall  TMainForm::ImportStructures(  AnsiString pFileName )
{
// Immediately following the 'Structures' keyword is a list of the different
// structures defined inside the file.
// By scanning rows for structure names we build a list of the various
// structures that we are going to parse.
// When a structure name is repeated it signifies that the first structure
// definition is following
// convert it as required
fStructureNames = new TStringList();
fStructureDescriptions = new TStringList();
fMemberNames = new TStringList();
fMemberTypes = new TStringList();
fMemberDescriptions = new TStringList();
fFlagNames = new TStringList();
fFlagDescriptions = new TStringList();

Dbg( ">>> determining the various structures included in this file <<<");
bool found_first_repeated_structure = false;
do
    {
    // check the current row to see if it contains a unique structure name
    Dbg( "checking row for a unique structure name" );
    if (IsStructureName(fStructureNames, ImportString))
        {
        Dbg( "found a repetition of a structure name: '" + ImportString + "'... assuming the relevant structure list is finished." );
        found_first_repeated_structure = true;
        break;  // continue at end of this for loop
        }

    // if we didn't find a unique structure name
    if (!found_first_repeated_structure)
        {
        // check that the row conforms to the structure naming rules
        if (IsValidStructureName(ImportString))
            {
            Dbg( "found a valid structure name: '" + ImportString + "'... adding to internal structures list" );
            fStructureNames->Add( ImportString );
            ImportRow++;
            if (ImportRow == ImportStrings->Count) ERROR( "scanned all rows but did not find a structure name repeated after the 'Structures included in this file' block...");
            }
        else ERROR( "scanned all rows but did not find a structure name repeated after the 'Structures included in this file' block...");
        }
    }
while (!found_first_repeated_structure);

// since we now know how many structures we need to parse, attempt to parse that many structures
for (int structure_index=0;structure_index<fStructureNames->Count;structure_index++)
    {
    Dbg( ">>> parsing structure definition #" + IntToStr(structure_index) + " <<<" );

    // clear internal storage
    TStringList* TempMemberNames = new TStringList();
    TStringList* TempMemberTypes = new TStringList();
    TStringList* TempMemberDescriptions = new TStringList();
    TStringList* TempFlagNames = new TStringList();
    TStringList* TempFlagDescriptions = new TStringList();

    // since we already know the names, check that the definition is appropriate but otherwise
    // ignore this row.
    Dbg( "Checking structure name to make sure that a definition is following");
    if (ImportString!=fStructureNames->Strings[structure_index]) ERROR( "Did not find the expected structure name: '" + fStructureNames->Strings[structure_index] + "'" );
    Dbg( "name is valid for structure: '" + fStructureNames->Strings[structure_index] + "'..." );
    Dbg( "... Assuming that a structure definition for '" + fStructureNames->Strings[structure_index] + "' is following" );
    Msg( "Parsing Structure: '" + fStructureNames->Strings[structure_index] + "'" );
    ImportRow++;

    // simply add the structure description
    Dbg( "Assuming row #" + IntToStr( ImportRow ) + " is a description for structure: '" + fStructureNames->Strings[structure_index] + "'" );
    fStructureDescriptions->Add( ImportString );
    ImportRow++;

    // by all rights we _must_ have a "struct" somewhere in this row
    Dbg( "checking row for a 'struct' construct..." );
    if (!IsStartOfStruct(ImportString)) ERROR( "There is no starting point for the 'struct' definition for the object '" + fStructureNames->Strings[structure_index] + "'" );
    Dbg( "... a 'struct' code declaration has been found..." );
    ImportRow++;

    // the following rows may occur anywhere.... hence we loop till we find the end of the 'struct' definition
    bool end_of_struct = false;
    int  inside_union_count = 0;
    Dbg( "parsing 'struct' definition..." );
    do
        {
        // skip out when we find the end-token
        if (IsEndOfStruct(ImportString))
            {
            Dbg( "Found the end of the 'struct' code declaration." );
            ImportRow++;
            end_of_struct = true;
            continue;
            }
        // ignore any 'union' declarations
        if (IsStartOfUnion(ImportString))
            {
            Dbg( "Found the start of a 'union' declaration... ignoring the 'union' row..." );
            ImportRow++;
            inside_union_count++;
            continue;
            }

        // ignore anything special while inside a 'union' declaration
        if (inside_union_count>0)
            {
            // ignore any '{' declarations
            if (ContainsStartBracket(ImportString))
                {
                Dbg( "ignoring row because it contains a '{' within a 'union' declaration" );
                ImportRow++;
                continue;
                }
            // ignore any '};' declarations  part of the 'union'
            if (IsEndOfUnion(ImportString))
                {
                Dbg( "Found the end of a 'union' declaration" );
                ImportRow++;
                inside_union_count--;
                continue;
                }
            }

        // determine the member type & name.
        
        // input:     LPRECT              lprDest;         // some comment
        // logic: LPRECT              lprDest;         // some comment
        // logic:       ^
        // result: LPRECT
        AnsiString temp_string = ImportString.TrimLeft();
        int temp_pos = temp_string.Pos(" ");
        AnsiString member_type = temp_string.SubString(1,temp_pos-1);
        Dbg( "found a structure member type: '" + member_type + "'" );
        TempMemberTypes->Add( member_type );
        // input: LPRECT              lprDest;         // some comment
        // logic: lprDest;         // some comment
        // logic:        ^
        // result: lprDest
        AnsiString member_name = temp_string.SubString(temp_pos, temp_string.Length() - temp_pos ).TrimLeft();
        temp_pos = member_name.Pos(";");
        if (temp_pos!=0) member_name = member_name.SubString( 1, temp_pos-1 );
        Dbg( "found a structure member name: '" + member_name + "'" );
        Msg( "Parsing Structure: '" + fStructureNames->Strings[structure_index] + "':: Found Member:'" + member_type + "  " + member_name + "'" );
        TempMemberNames->Add( member_name );

        ImportRow++;
        if (ImportRow == ImportStrings->Count) ERROR( "scanned all rows but did not find the end of the 'struct' declaration.");
        }
    while (!end_of_struct);

    // we now expect the word "Members"... if not something has gone wrong.
    Dbg( "expecting a 'Members' keyword... comparing..." );
    if (!IsKeyword(ImportString,"Members")) ERROR( "Did not find the 'Members' keyword as expected." );
    Dbg( "row contained a valid 'Members' keyword" );
    ImportRow++;

    // we can now parse the descriptions of the various members of the structure.
    // note that some members also contain descriptions of various flags.
    // note also that other members contain "something and somethingelse" which
    // we need to distinguish between.
    Dbg( "attempting to parse the descriptions for the various structure members" );
    bool checking_last_member_flags = false;
    for (int member_index=0;member_index<TempMemberNames->Count;member_index++)
        {
        Dbg( "parsing structure member #" + IntToStr(member_index) );
        // check for the premature end of the members descriptions
        if (IsKeyword(ImportString,"QuickInfo"))
            {
            // special case: if we're checking for flags for the last member, and the check failed
            // do not check for "this and that" rules.
            if (!checking_last_member_flags) ERROR( "found a 'QuickInfo' section before all member descriptions were defined." );
            Dbg( "found a 'QuickInfo' section after checking all member descriptions. Assuming member descriptions are completed" );
            break;
            }
        if (IsKeyword(ImportString,"Remarks"))
            {
            // special case: if we're checking for flags for the last member, and the check failed
            // do not check for "this and that" rules.
            if (!checking_last_member_flags) ERROR( "found a 'Remarks' section before all member descriptions were defined." );
            Dbg( "found a 'Remarks' section after checking all member descriptions. Assuming member descriptions are completed" );
            break;
            }

        // if the current line contains the name of the appropriate member, ignore the member name &
        // store the description that follows immediately after this row.
        Dbg( "expecting member name: '" + TempMemberNames->Strings[member_index] + "'... checking to make sure that row is a valid member name" );
        if (ImportString==TempMemberNames->Strings[member_index])
            {
            Dbg( "found a valid member name: '" + TempMemberNames->Strings[member_index] + "'" );
            Msg( "Parsing Structure Member: '" + fStructureNames->Strings[structure_index] + "::" + TempMemberNames->Strings[member_index] + "'" );
            ImportRow++;

            Dbg( "assuming row #" + IntToStr(ImportRow) + " as the description for member: '" + TempMemberNames->Strings[member_index] + "'" );
            TempMemberDescriptions->Add( EncodeValue("StructureIndex",structure_index) + ImportString );
            ImportRow++;

            // check if there are any flags following the member description
            Dbg( "checking to see if flags are declared following the member description" );
            if (IsFlag(ImportString))
                {
                Dbg( "found a flag rule immdediately following the member description." );
                do
                    {
                    Dbg( "adding a member flag: '" + ImportString + "'" );
                    Msg( "Found Structure Member Flag: '" + fStructureNames->Strings[structure_index] + "::" + TempMemberNames->Strings[member_index] + "'" + "::" + ImportString );
                    TempFlagNames->Add( ImportString);
                    // check for flag descriptions declared on the same line as the flag name
                    int temp_pos = ImportString.Pos(" ");
                    if (temp_pos!=0)
                        {
                        Dbg( "the member flag row also contains the relevant flag description" );
                        AnsiString flag_description = ImportString.SubString(temp_pos,ImportString.Length()-temp_pos).TrimLeft();
                        Dbg( "adding the member flag description: '" + flag_description + "'" );
                        TempFlagDescriptions->Add(EncodeValue("FlagStructureIndex",structure_index) + EncodeValue("FlagMemberIndex",member_index) + flag_description);
                        }
                    else
                        // check just to make sure that the flags do not have descriptions with them...
                        if (IsFlag(ImportString))
                            {
                            Dbg( "the member flag description is assumed to be on the next row..." );
                            ImportRow++;
                            Dbg("adding a member flag description: '" + ImportString + "'" );
                            TempFlagDescriptions->Add(EncodeValue("FlagStructureIndex",structure_index) + EncodeValue("FlagMemberIndex",member_index) + ImportString);
                            ImportRow++;
                            }
                        else
                            ERROR( "a flag was encountered followed immediately by another flag" );
                    }
                while (IsFlag(ImportString));
                }
            else
                Dbg( "there are no flags following the member description" );
            }
        else
            // it's not a simple member.... maybe it's a case of "this and that".... or "this, that, and something else"
            {
            Dbg( "did not find a member name where expected... checking 'this and that' rules..." );
            // if there is another member name following this one....
            if (member_index<TempMemberNames->Count-1)
                {
                Dbg( "...checking 'this and that' rules");
                AnsiString first_member_name;
                AnsiString second_member_name;
                if (IsThisAndThat(TempMemberNames,ImportString,&first_member_name,&second_member_name))
                    {
                    Dbg( "the row fits the rules for a 'this and that' type description" );
                    ImportRow++;    // skip past the "this and that" row
                    Dbg( "assuming row #" + IntToStr(ImportRow) + " is the description for members: '" + first_member_name + "' and '" + second_member_name + "'" );
                    TempMemberDescriptions->Add( EncodeValue("StructureIndex",structure_index) + ImportString );
                    TempMemberDescriptions->Add( EncodeValue("StructureIndex",structure_index) + ImportString );
                    ImportRow++;
                    member_index++;
                    }
                else
                    {
                    Dbg( "the row did not fit the 'this and that' rule. Checking 'this, that, and something else' rule..." );
                    TStringList* temp_member_names = new TStringList();
                    if (IsThisAndThatAndSomethingElse(TempMemberNames,ImportString, temp_member_names ))
                        {
                        Dbg( "the row fits the rules for a 'this, and that and something else' type description" );
                        ImportRow++;    // skip past the "this, that and something else" row
                        for (int i=0;i<temp_member_names->Count;i++)
                            {
                            Dbg( "adding member description for member: '" + temp_member_names->Strings[i] + "'" ); 
                            TempMemberDescriptions->Add( EncodeValue("StructureIndex",structure_index) + ImportString );
                            }
                        if (temp_member_names->Count>0) member_index += temp_member_names->Count-1;
                        ImportRow++;
                        }
                    else
                        {
                        Dbg( "Found the following member names when checking 'this, that, and something else' rule" );
                        for (int i=0;i<temp_member_names->Count;i++) Dbg( "member #" + IntToStr(i) + "=='" + temp_member_names->Strings[i] + "'" );
                        temp_member_names->Clear();
                        delete temp_member_names;
                        ERROR( "Expected but did not find a particular member name & description: '" + fStructureNames->Strings[structure_index] + "::" + TempMemberNames->Strings[member_index] + "'" );
                        }
                    temp_member_names->Clear();
                    delete temp_member_names;
                    }
                }
            else
                ERROR( "Expected but did not find a particular member name & description: '" + fStructureNames->Strings[structure_index] + "::" + TempMemberNames->Strings[member_index] + "'" );
            }
        }

    // we now expect either a "Remarks" or a "QuickInfo" section
    // check just to make sure
    // if it is a 'Remarks' section, simply skip it till we find a "QuickInfo" section.

    if (ImportString=="Remarks")
        while (ImportString!="QuickInfo")
            {
            ImportRow++;
            if (ImportRow>=ImportStrings->Count)
                {
                Dbg( "Could not parse the import file '" + pFileName + "'.\n" +
                               "Found a 'Remarks' section but did not find the 'QuickInfo' section following" +
                                ImportString );
                Dbg( "ERROR LINE: " + ImportString );
                DebuggingForm->Show();
                DebugLabel->Caption = "";
                return;
                }
            }

//    Dbg( "Checking if the keyword 'QuickInfo' exists.");
//    Dbg( "string being checked == '" + ImportString + "'" );
    if (ImportString!="QuickInfo")
        {
        Dbg( "Could not parse the import file '" + pFileName + "'.\n" +
                    "Could not find a 'QuickInfo' section" );
        Dbg( "ERROR LINE: " + ImportString );
        DebuggingForm->Show();
        DebugLabel->Caption = "";
        return;
        }

    // skip past the four lines
//    Dbg( "The keyword 'QuickInfo' exists. skipping the 'QuickInfo' block");
    ImportRow += 5;

    // yey. the structure definition has been successfully parsed.
    Dbg( ">>>>>> Definition of structure '" + fStructureNames->Strings[structure_index] + "' was completed successfully! <<<<<<" );

    //    Dbg( "Clearing structure internal representations" );
    int i;
    for (i=0;i<TempFlagDescriptions->Count;i++) fFlagDescriptions->Add( TempFlagDescriptions->Strings[i] );
    for (i=0;i<TempFlagNames->Count;i++) fFlagNames->Add( TempFlagNames->Strings[i] );
    for (i=0;i<TempMemberDescriptions->Count;i++) fMemberDescriptions->Add( TempMemberDescriptions->Strings[i] );
    for (i=0;i<TempMemberTypes->Count;i++) fMemberTypes->Add( TempMemberTypes->Strings[i] );
    for (i=0;i<TempMemberNames->Count;i++) fMemberNames->Add( TempMemberNames->Strings[i] );

    TempFlagDescriptions->Clear();
    TempFlagNames->Clear();
    TempMemberDescriptions->Clear();
    TempMemberTypes->Clear();
    TempMemberNames->Clear();

    delete TempFlagDescriptions;
    delete TempFlagNames;
    delete TempMemberDescriptions;
    delete TempMemberTypes;
    delete TempMemberNames;
    }

Dbg( IntToStr(fStructureNames->Count) + " structures were imported SUCCESSFULLY!" );
Msg( IntToStr(fStructureNames->Count) + " structures were imported SUCCESSFULLY!" );

Dbg("");
Dbg( "Structures DUMP :-" );
Msg( "Dumping structures to database" );
Dbg("");
int structure;
int member;
int flag;
Dbg( "# of structures names == " + IntToStr(fStructureNames->Count) );
Dbg( "# of structures descriptions == " + IntToStr(fStructureDescriptions->Count) );
Dbg( "# of member names == " + IntToStr(fMemberNames->Count) );
Dbg( "# of member types == " + IntToStr(fMemberTypes->Count) );
Dbg( "# of member descriptions == " + IntToStr(fMemberDescriptions->Count) );
Dbg( "# of flag names == " + IntToStr(fFlagNames->Count) );
Dbg( "# of flag descriptions == " + IntToStr(fFlagDescriptions->Count) );

Structures_Table->Close();       // close and re-open so that we know for sure that the table has been opened
Structures_Table->Open();
for (structure=0;structure<fStructureNames->Count;structure++)
    {
    Dbg( "Saving Structure #" + IntToStr(structure) + ": '" + fStructureNames->Strings[structure] + "'" );
    Msg( "Saving Structure #" + IntToStr(structure) + ": '" + fStructureNames->Strings[structure] + "'" );
    Structures_Table->Append();
    Structures_Table->FieldByName("Conversion_Status")->AsInteger = 0;
    Structures_Table->FieldByName("Foundation_Class_Index")->AsInteger = fFoundation_Class_Index;
    Structures_Table->FieldByName("Structure_Name")->AsString = fStructureNames->Strings[structure];
    Structures_Table->FieldByName("Structure_Description")->AsString = fStructureDescriptions->Strings[structure];
    Structures_Table->FieldByName("BCB_Name")->AsString = fStructureNames->Strings[structure].LowerCase();
    Structures_Table->FieldByName("BCB_Description")->AsString = fStructureDescriptions->Strings[structure];
    Structures_Table->Post();
    // retrieve & store the unique field value (created during the Post() but not yet
    Structures_Table->Refresh();
//    TempQuery->Close();
//    TempQuery->SQL->Clear();
//    TempQuery->SQL->Add( " select Structure_Index from Structures.DB" );
//    TempQuery->SQL->Add( " where Structure_Name = '" + fStructureNames->Strings[structure] + "'" );
//    TempQuery->Open();
//    Dbg( "TempQuery:: Structure #" + IntToStr(structure) + " [" + fStructureNames->Strings[structure] + "] unique index == " + TempQuery->FieldByName("Structure_Index")->AsString );
//    TempQuery->Close();

//    Dbg( "Structures_Table:: Structure #" + IntToStr(structure) + " [" + fStructureNames->Strings[structure] + "] unique index == " + Structures_Table->FieldByName("Structure_Index")->AsString );

//    Dbg( "original structure name == '" + fStructureNames->Strings[structure] + "'" );
    fStructureNames->Strings[structure] = EncodeValue( "UniqueStructureIndex", Structures_Table->FieldByName("Structure_Index")->AsInteger) + fStructureNames->Strings[structure];
//    TempQuery->Close();
//    Dbg( "modified structure name with unique field == '" + fStructureNames->Strings[structure] + "'" );
    }

StructureMembers_Table->Close();
StructureMembers_Table->Open();
AnsiString decoded_member_description;
AnsiString crap_to_be_ignored;
for (member=0;member<fMemberNames->Count;member++)
    {
    structure = DecodeValue( "StructureIndex", fMemberDescriptions->Strings[member], &decoded_member_description );
    int unique_structure_index = DecodeValue( "UniqueStructureIndex", fStructureNames->Strings[structure], &crap_to_be_ignored );

    Dbg( "Saving Structure Member #" + IntToStr(member) + ": '" + fStructureNames->Strings[structure] + "::" + fMemberNames->Strings[member] + "'" );
    Msg( "Saving Structure Member #" + IntToStr(member) + ": '" + fStructureNames->Strings[structure] + "::" + fMemberNames->Strings[member] + "'" );

    StructureMembers_Table->Append();
    StructureMembers_Table->FieldByName("Conversion_Status")->AsInteger = 0;
    StructureMembers_Table->FieldByName("Foundation_Class_Index")->AsInteger = fFoundation_Class_Index;
    StructureMembers_Table->FieldByName("Structure_Index")->AsInteger = unique_structure_index;
    StructureMembers_Table->FieldByName("Member_Name")->AsString = fMemberNames->Strings[member];
    StructureMembers_Table->FieldByName("Member_Type")->AsString = fMemberTypes->Strings[member];
    StructureMembers_Table->FieldByName("Member_Description")->AsString= decoded_member_description;
    StructureMembers_Table->FieldByName("BCB_Name")->AsString = RemoveHungarianNotation( fMemberNames->Strings[member] );
    StructureMembers_Table->FieldByName("BCB_Description")->AsString= decoded_member_description;
    StructureMembers_Table->Post();
    StructureMembers_Table->Refresh();

//    TempQuery->Close();
//    TempQuery->SQL->Clear();
//    TempQuery->SQL->Add( " select Member_Index from Structures.DB" );
//    TempQuery->SQL->Add( " where Structure_Name = '" + fStructureNames->Strings[structure] + "'" );
//    TempQuery->Open();
//    fStructureNames->Strings[structure] = EncodeValue( "UniqueStructureIndex", Structures_Table->FieldByName("Foundation_Class_Index")->AsInteger) + fStructureNames->Strings[structure];
//    TempQuery->Close();

//    Dbg( "original member name == '" + fStructureNames->Strings[structure] + "'" );
    fMemberNames->Strings[member] = EncodeValue( "UniqueMemberIndex", StructureMembers_Table->FieldByName("Member_Index")->AsInteger) + fMemberNames->Strings[member];
//    Dbg( "modified member name with unique field == '" + fMemberNames->Strings[member] + "'" );
    }

StructureMemberFlags_Table->Close();
StructureMemberFlags_Table->Open();
AnsiString decoded_flag_description;
for (flag=0;flag<fFlagNames->Count;flag++)
    {
    structure = DecodeValue( "FlagStructureIndex", fFlagDescriptions->Strings[flag], &decoded_flag_description );
    int unique_structure_index = DecodeValue( "UniqueStructureIndex", fStructureNames->Strings[structure], &crap_to_be_ignored );

    member = DecodeValue( "FlagMemberIndex", fFlagDescriptions->Strings[flag], &decoded_flag_description );
    int unique_member_index = DecodeValue( "UniqueMemberIndex", fMemberNames->Strings[member], &crap_to_be_ignored );

    Dbg( "Saving Structure Member Flag #" + IntToStr(flag) + ": '" + fStructureNames->Strings[structure] + "::" + fMemberNames->Strings[member] + "::" + fFlagNames->Strings[flag] + "'" );
    Msg( "Saving Structure Member Flag #" + IntToStr(flag) + ": '" + fStructureNames->Strings[structure] + "::" + fMemberNames->Strings[member] + "::" + fFlagNames->Strings[flag] + "'" );

    StructureMemberFlags_Table->Append();
    StructureMemberFlags_Table->FieldByName("Conversion_Status")->AsInteger = 0;
    StructureMemberFlags_Table->FieldByName("Foundation_Class_Index")->AsInteger = fFoundation_Class_Index;
    StructureMemberFlags_Table->FieldByName("Structure_Index")->AsInteger = unique_structure_index;
    StructureMemberFlags_Table->FieldByName("Member_Index")->AsInteger = unique_member_index;
    StructureMemberFlags_Table->FieldByName("Flag_Name")->AsString = fFlagNames->Strings[flag];
    StructureMemberFlags_Table->FieldByName("Flag_Description")->AsString = decoded_flag_description;
    StructureMemberFlags_Table->FieldByName("BCB_Description")->AsString = decoded_flag_description;
    StructureMemberFlags_Table->Post();
    }

// clean up usage
fFlagDescriptions->Clear();
fFlagNames->Clear();
fMemberDescriptions->Clear();
fMemberTypes->Clear();
fMemberNames->Clear();
fStructureDescriptions->Clear();
fStructureNames->Clear();

Dbg( "Done. All structures were imported successfully!" );
Msg( "Done. All structures were imported successfully!" );

delete fFlagDescriptions;
delete fFlagNames;
delete fMemberDescriptions;
delete fMemberTypes;
delete fMemberNames;
delete fStructureDescriptions;
delete fStructureNames;
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsStructureName( TStringList* structure_name_strings, AnsiString structure_name )
{
for (int i=0;i<structure_name_strings->Count;i++) if (structure_name_strings->Strings[i]==structure_name) return true;
return false;
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsMemberName( TStringList* member_name_strings, AnsiString member_name )
{
for (int i=0;i<member_name_strings->Count;i++) if (member_name_strings->Strings[i]==member_name) return true;
return false;
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::BuildComboSelections()
{
BuildStructureSelections();
BuildMemberSelections();
BuildFlagSelections();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BuildStructureSelections()
{
// build a list of current known structures & put them in the DirectX_ComboBox & BCB_ComboBox
TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add(" select Structure_Name, BCB_Name from Structures.DB s, Foundation_Classes.DB f " );
TempQuery->SQL->Add(" where s.Foundation_Class_Index = f.Foundation_Class_Index " );
TempQuery->SQL->Add(" and f.Foundation_Class_Name = '" + DirectXFoundationClassPages->ActivePage->Caption + "'" );
TempQuery->SQL->Add(" order by Structure_Name ");
TempQuery->Open();
TempQuery->First();
int orig_itemindex = DirectX_ComboBox->ItemIndex;
if (fLinked)
    {
    DirectX_ComboBox->Items->Clear();
    BCB_ComboBox->Items->Clear();
    }
else
    if (DirectDrawEditPages->ActivePage->Caption == "DirectX") DirectX_ComboBox->Items->Clear();
    else BCB_ComboBox->Items->Clear();
while (!TempQuery->Eof)
    {
    if (fLinked)
        {
        DirectX_ComboBox->Items->Add(TempQuery->FieldByName("Structure_Name")->AsString);
        BCB_ComboBox->Items->Add(TempQuery->FieldByName("BCB_Name")->AsString);
        }
    else
        if (DirectDrawEditPages->ActivePage->Caption == "DirectX") DirectX_ComboBox->Items->Add(TempQuery->FieldByName("Structure_Name")->AsString);
        else DirectX_ComboBox->Items->Add(TempQuery->FieldByName("Structure_Name")->AsString);
    TempQuery->Next();
    }
TempQuery->Close();
if (fLinked)
    {
    DirectX_ComboBox->ItemIndex = orig_itemindex;
    BCB_ComboBox->ItemIndex = orig_itemindex;
    }
else
    if (DirectDrawEditPages->ActivePage->Caption == "DirectX") DirectX_ComboBox->ItemIndex = orig_itemindex;
    else BCB_ComboBox->ItemIndex = orig_itemindex;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BuildMemberSelections()
{
// fill in the list of known structure members & put them in the DirectX_StructureMember_ComboBox & BCB_ObjectProperty_ComboBox
TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add(" select Member_Name, BCB_Name from Structure_Members.DB" );
TempQuery->SQL->Add(" where Structure_Index = " + IntToStr( fStructure_Index ) );
TempQuery->SQL->Add(" order by Member_Name ");
TempQuery->Open();
TempQuery->First();
int orig_itemindex = DirectX_StructureMember_ComboBox->ItemIndex;
if (fLinked)
    {
    DirectX_StructureMember_ComboBox->Items->Clear();
    BCB_ObjectProperty_ComboBox->Items->Clear();
    }
else
    if (DirectDrawEditPages->ActivePage->Caption == "DirectX") DirectX_StructureMember_ComboBox->Items->Clear();
    else BCB_ObjectProperty_ComboBox->Items->Clear();
while (!TempQuery->Eof)
    {
    if (fLinked)
        {
        DirectX_StructureMember_ComboBox->Items->Add(TempQuery->FieldByName("Member_Name")->AsString);
        BCB_ObjectProperty_ComboBox->Items->Add(TempQuery->FieldByName("BCB_Name")->AsString);
        }
    else
        if (DirectDrawEditPages->ActivePage->Caption == "DirectX") DirectX_StructureMember_ComboBox->Items->Add(TempQuery->FieldByName("Member_Name")->AsString);
        else BCB_ObjectProperty_ComboBox->Items->Add(TempQuery->FieldByName("BCB_Name")->AsString);
    TempQuery->Next();
    }
TempQuery->Close();
if (fLinked)
    {
    DirectX_StructureMember_ComboBox->ItemIndex = orig_itemindex;
    BCB_ObjectProperty_ComboBox->ItemIndex = orig_itemindex;
    }
else
    if (DirectDrawEditPages->ActivePage->Caption == "DirectX") DirectX_StructureMember_ComboBox->ItemIndex = orig_itemindex;
    else BCB_ObjectProperty_ComboBox->ItemIndex = orig_itemindex;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BuildFlagSelections()
{
// fill in the list of known member flags & put them in the DirectX_MemberFlags_ComboBox & BCB_PropertyFlags_ComboBox
TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add(" select Flag_Name from Structure_Member_Flags.DB " );
TempQuery->SQL->Add(" where Member_Index = " + IntToStr( fMember_Index ) );
TempQuery->SQL->Add(" order by Flag_Name ");
TempQuery->Open();
TempQuery->First();
int orig_itemindex = DirectX_MemberFlags_ComboBox->ItemIndex;
if (fLinked)
    {
    DirectX_MemberFlags_ComboBox->Items->Clear();
    BCB_PropertyFlags_ComboBox->Items->Clear();
    }
else
    if (DirectDrawEditPages->ActivePage->Caption == "DirectX") DirectX_MemberFlags_ComboBox->Items->Clear();
    else BCB_PropertyFlags_ComboBox->Items->Clear();
while (!TempQuery->Eof)
    {
    if (fLinked)
        {
        DirectX_MemberFlags_ComboBox->Items->Add(TempQuery->FieldByName("Flag_Name")->AsString);
        BCB_PropertyFlags_ComboBox->Items->Add(TempQuery->FieldByName("Flag_Name")->AsString);
        }
    else
        if (DirectDrawEditPages->ActivePage->Caption == "DirectX") DirectX_MemberFlags_ComboBox->Items->Add(TempQuery->FieldByName("Flag_Name")->AsString);
        else BCB_PropertyFlags_ComboBox->Items->Add(TempQuery->FieldByName("Flag_Name")->AsString);
    TempQuery->Next();
    }
TempQuery->Close();
if (fLinked)
    {
    DirectX_MemberFlags_ComboBox->ItemIndex = orig_itemindex;
    BCB_PropertyFlags_ComboBox->ItemIndex = orig_itemindex;
    }
else
    if (DirectDrawEditPages->ActivePage->Caption == "DirectX") DirectX_MemberFlags_ComboBox->ItemIndex = orig_itemindex;
    else BCB_PropertyFlags_ComboBox->ItemIndex = orig_itemindex;
//DirectX_MemberFlags_GroupBox->Visible = (DirectX_MemberFlags_ComboBox->Items->Count!=0);
//BCB_ObjectFlags_GroupBox->Visible = (DirectX_MemberFlags_ComboBox->Items->Count!=0);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormActivate(TObject *Sender)
{
DirectX_To_BCB->Connected = true;
Structures_Table->Active = true;
StructureMembers_Table->Active = true;
StructureMemberFlags_Table->Active = true;
if (Structures_Table->RecordCount==0) Structures_Table->Insert(); else Structures_Table->Edit();
if (StructureMembers_Table->RecordCount==0) StructureMembers_Table->Insert(); else StructureMembers_Table->Edit();
if (StructureMemberFlags_Table->RecordCount==0) StructureMemberFlags_Table->Insert(); else StructureMemberFlags_Table->Edit();

// force the main pages to change so that the OnChange event occurs
DirectXFoundationClassPagesChange(Sender);
// force the combo boxes to contain valid information
BuildComboSelections();

DirectX_ComboBox->ItemIndex = 0;
DirectX_StructureMember_ComboBox->ItemIndex = 0;
DirectX_MemberFlags_ComboBox->ItemIndex = 0;
BCB_ComboBox->ItemIndex = 0;
BCB_ObjectProperty_ComboBox->ItemIndex = 0;
BCB_PropertyFlags_ComboBox->ItemIndex = 0;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormDeactivate(TObject *Sender)
{
Structures_Table->Active = false;
StructureMembers_Table->Active = false;
StructureMemberFlags_Table->Active = false;
DirectX_To_BCB->Connected = false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
Top = 0;
Left = 0;
Width = 800;
Height = 580;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BitBtn3Click(TObject *Sender)
{
ShowMessage("NOTE: Clearing _ALL_ Structures, Methods, Flags, etc.");
Structures_Table->Close();
Structures_Table->EmptyTable();
Structures_Table->Open();
StructureMembers_Table->Close();
StructureMembers_Table->EmptyTable();
StructureMembers_Table->Open();
StructureMemberFlags_Table->Close();
StructureMemberFlags_Table->EmptyTable();
StructureMemberFlags_Table->Open();

BuildComboSelections();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BCB_ObjectName_EditExit(TObject *Sender)
{
if (Structures_Table->Modified)
    {
    Structures_Table->Post();
    BuildStructureSelections();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BCB_PropertyName_EditExit(
    TObject *Sender)
{
if (StructureMembers_Table->Modified)
    {
    StructureMembers_Table->Post();
    BuildMemberSelections();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BCB_PropertyFlagsDescript_MemoExit(
    TObject *Sender)
{
if (StructureMemberFlags_Table->Modified)
    {
    StructureMemberFlags_Table->Post();
    BuildFlagSelections();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DBRadioGroup1Exit(TObject *Sender)
{
if (Structures_Table->Modified) Structures_Table->Post();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BitBtn4Click(TObject *Sender)
{
DebuggingForm->Visible = !DebuggingForm->Visible;    
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Label7Click(TObject *Sender)
{
BCB_ComboBox->SetFocus();    
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label6Click(TObject *Sender)
{
BCB_ObjectName_Edit->SetFocus();    
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label5Click(TObject *Sender)
{
BCB_ObjectDescription_Memo->SetFocus();    
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label9Click(TObject *Sender)
{
BCB_ObjectProperty_ComboBox->SetFocus();    
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label14Click(TObject *Sender)
{
BCB_PropertyName_Edit->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label15Click(TObject *Sender)
{
BCB_PropertyDescription_Memo->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label16Click(TObject *Sender)
{
BCB_PropertyFlags_ComboBox->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label17Click(TObject *Sender)
{
BCB_PropertyFlagsDescript_Memo->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BitBtn7Click(TObject *Sender)
{
if (ImportFileDialog->Execute()) ImportFile( ImportFileDialog->FileName );
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DirectXFoundationClassPagesChange(TObject *Sender)
{
TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add( " select Foundation_Class_Index from Foundation_Classes.DB " );
TempQuery->SQL->Add( " where Foundation_Class_Name = '" + DirectXFoundationClassPages->ActivePage->Caption + "'" );
TempQuery->Open();
TempQuery->First();
fFoundation_Class_Index = TempQuery->FieldByName("Foundation_Class_Index")->AsInteger;
TempQuery->Close();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DirectX_ComboBoxChange(TObject *Sender)
{
if (DirectX_ComboBox->ItemIndex!=-1)
    {
//    BCB_ComboBox->ItemIndex = DirectX_ComboBox->ItemIndex;

    // note: seems a bit wierd to use this query here.... but ....
    // even though the Structure_Name field is declared as an index (using Database Desktop) &
    // the Structure_Name field is actually determined as an index by the object inspector, when we
    // attempt to use it as a search key for GotoKey(), etc. BCB responds
    // "there is no key field called Structure_Name" ... huh?
    // hence we retrieve an use a valid index field manually.

    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Structure_Index from Structures.DB " );
    TempQuery->SQL->Add(" where Structure_Name = '" + DirectX_ComboBox->Items->Strings[DirectX_ComboBox->ItemIndex] + "'" );
    TempQuery->Open();
    fStructure_Index = TempQuery->FieldByName("Structure_Index")->AsInteger;

    TempQuery->Close();

    Structures_Table->Close();
    Structures_Table->IndexFieldNames = "Structure_Index";
    Structures_Table->Open ();

    Structures_Table->SetKey();
    Structures_Table->FieldByName("Structure_Index")->AsInteger = fStructure_Index;
    Structures_Table->GotoNearest();

    BuildMemberSelections();
    DirectX_StructureMember_ComboBoxChange(Sender);
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DirectX_StructureMember_ComboBoxChange(
    TObject *Sender)
{
if ((DirectX_ComboBox->ItemIndex!=-1) &&
    (DirectX_StructureMember_ComboBox->ItemIndex!=-1))
    {
    BCB_ObjectProperty_ComboBox->ItemIndex = DirectX_StructureMember_ComboBox->ItemIndex;

    // note: seems a bit wierd to use this query here.... but ....
    // even though the Structure_Name field is declared as an index (using Database Desktop) &
    // the Structure_Name field is actually determined as an index by the object inspector, when we
    // attempt to use it as a search key for GotoKey(), etc. BCB responds
    // "there is no key field called Structure_Name" ... huh?
    // hence we retrieve an use a valid index field manually.

    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Member_Index from Structure_Members.DB " );
    TempQuery->SQL->Add(" where Member_Name = '" + DirectX_StructureMember_ComboBox->Items->Strings[DirectX_StructureMember_ComboBox->ItemIndex] + "'" );
    TempQuery->Open();
    fMember_Index = TempQuery->FieldByName("Member_Index")->AsInteger;

    TempQuery->Close();

    StructureMembers_Table->Close();
    StructureMembers_Table->IndexFieldNames = "Member_Index";
    StructureMembers_Table->Open ();

    StructureMembers_Table->SetKey();
    StructureMembers_Table->FieldByName("Member_Index")->AsInteger = fMember_Index;
    StructureMembers_Table->GotoNearest();

    BuildFlagSelections();
    DirectX_MemberFlags_ComboBoxChange(Sender);
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DirectX_MemberFlags_ComboBoxChange(
    TObject *Sender)
{
if ((DirectX_ComboBox->ItemIndex!=-1) &&
    (DirectX_StructureMember_ComboBox->ItemIndex!=-1) &&
    (DirectX_MemberFlags_ComboBox->ItemIndex!=-1))
    {
    BCB_PropertyFlags_ComboBox->ItemIndex = DirectX_MemberFlags_ComboBox->ItemIndex;
    // note: seems a bit wierd to use this query here.... but ....
    // even though the Structure_Name field is declared as an index (using Database Desktop) &
    // the Structure_Name field is actually determined as an index by the object inspector, when we
    // attempt to use it as a search key for GotoKey(), etc. BCB responds
    // "there is no key field called Structure_Name" ... huh?
    // hence we retrieve an use a valid index field manually.

    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Flag_Index from Structure_Member_Flags.DB " );
    TempQuery->SQL->Add(" where Flag_Name = '" + DirectX_MemberFlags_ComboBox->Items->Strings[DirectX_MemberFlags_ComboBox->ItemIndex] + "'" );
    TempQuery->Open();
    int flag_index = TempQuery->FieldByName("Flag_Index")->AsInteger;

    TempQuery->Close();

    StructureMemberFlags_Table->Close();
    StructureMemberFlags_Table->IndexFieldNames = "Flag_Index";
    StructureMemberFlags_Table->Open ();

    StructureMemberFlags_Table->SetKey();
    StructureMemberFlags_Table->FieldByName("Flag_Index")->AsInteger = flag_index;
    StructureMemberFlags_Table->GotoNearest();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BitBtn8Click(TObject *Sender)
{
fLinked = true;
LinkedLabel->Caption = "Linked: " + (fLinked) ? "True" : "False";
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BitBtn9Click(TObject *Sender)
{
fLinked = false;
LinkedLabel->Caption = "Linked: " + (fLinked) ? "True" : "False";
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BitBtn6Click(TObject *Sender)
{
ShowMessage("Not Yet Implemented");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BitBtn10Click(TObject *Sender)
{
ShowMessage("Not Yet Implemented");
    
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BitBtn5Click(TObject *Sender)
{
ShowMessage("Not Yet Implemented");
    
}
//---------------------------------------------------------------------------
